/* $Header:   F:/source/cwright/addons/sc/sc.c_v   1.27   27 Aug 1992 12:23:22   ericj  $ */
/*
 *
 *                  Copyright 1991,1992 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 *
 *
 * PG.C  Adept V+ language extensions for CodeWright V2.0
 * Adept V+ Version 10.3
 *
 * 01/31/93  DKT  Initial Code
 *
 *
 *
 * Notes:
 *   The code in this file is derived from the ..\cwright\sc\sc.c file
 *   provided by Premia.
 *
 * TODO:
 *
 *   - The '+' or '-' in binary, hex, and octal numbers is being included
 *     in the digit count, and it shouldn't be.
 *     (ex. ^HFFFFFFFF is colored, but ^H-FFFFFFFF is not colored)
 *
 * Copyright 1993, Vanguard Automation, Inc.  All rights reserved.
 *
 */

/*#include "win_filt.h"*/
#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

#define MAXDIGIT1 9   /* max # digits (minus one) in an integer value (i.e. ^H1F, ^B101) */
#define MAXDIGIT2 10  /* max # digits (minus one) in a real value (i.e. 47, 10.902, 34.7E+2) */
#define MAXDIGIT3 3   /* max # digits (minus one) after the exponent in a real value */

LIBMAIN



#define MAX_BUF_LEN    100

static LPVOID _pg_hash_names        = NIL;  /* pointer to the hash table containing */
                                            /* all 'Adept V+' keyword names. */

static LPVOID _pg_hash_templates    = NIL;  /* pointer to the hash table containing */
					   /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _pg_hash_templates.
 */
typedef struct _LanguageTemplate {
	LPSTR abbrev;
	LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
#define PG_INDENT1 4
#define PG_INDENT2 2

LPSTR PG_INDENT_STR1 = "\377IF\377ELSE\377DO\377FOR\377WHILE\377";
LPSTR PG_INDENT_STR2 = "\377CASE\377VALUE\377ANY\377";

/*
 * _pg_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _pg_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */

LanguageTemplate _pg_templates[] = {
	{ ".P",         ".PROGRAM &()\n.END" },
	{ ".PR",        ".PROGRAM &()\n.END" },
	{ ".PRO",       ".PROGRAM &()\n.END" },
	{ ".PROG",      ".PROGRAM &()\n.END" },
	{ ".PROGR",     ".PROGRAM &()\n.END" },
	{ ".PROGRA",    ".PROGRAM &()\n.END" },
	{ ".PROGRAM",   ".PROGRAM &()\n.END" },
	{ "ANY",        "ANY", },
	{ "A",          "AUTO ", },
	{ "AU",         "AUTO ", },
	{ "AUT",        "AUTO ", },
	{ "AUTO",       "AUTO ", },
	{ "AR",         "AUTO REAL " },
	{ "AL",         "AUTO LOC " },
	{ "AP",         "APPRO &, " },
	{ "APP",        "APPRO &, " },
	{ "APPR",       "APPRO &, " },
	{ "APPRO",      "APPRO &, " },
	{ "APB",        "APPRO &, \nBREAK" },
	{ "APPB",       "APPRO &, \nBREAK" },
	{ "APPRB",      "APPRO &, \nBREAK" },
	{ "APPROB",     "APPRO &, \nBREAK" },
	{ "APS",        "APPROS &, " },
	{ "APPS",       "APPROS &, " },
	{ "APPRS",      "APPROS &, " },
	{ "APPROS",     "APPROS &, " },
	{ "APSB",       "APPROS &, \nBREAK" },
	{ "APPSB",      "APPROS &, \nBREAK" },
	{ "APPRSB",     "APPROS &, \nBREAK" },
	{ "APPROSB",    "APPROS &, \nBREAK" },
	{ "AT",         "ATTACH (&)" },
	{ "ATT",        "ATTACH (&)" },
	{ "ATTA",       "ATTACH (&)" },
	{ "ATTAC",      "ATTACH (&)" },
	{ "ATTACH",     "ATTACH (&)" },
	{ "BR",         "BREAK" },
	{ "BRE",        "BREAK" },
	{ "BREA",       "BREAK" },
	{ "BREAK",      "BREAK" },
	{ "CALL",       "CALL &()" },
	{ "CAS",        "CASE & OF\nEND" },
	{ "CASE",       "CASE & OF\nEND" },
	{ "CAS;",       "CASE & OF\nEND ; case" },
	{ "CASE;",      "CASE & OF\nEND ; case" },
	{ "DEC",        "DECOMPOSE &[] = " },
	{ "DECO",       "DECOMPOSE &[] = " },
	{ "DECOM",      "DECOMPOSE &[] = " },
	{ "DECOMP",     "DECOMPOSE &[] = " },
	{ "DECOMPO",    "DECOMPOSE &[] = " },
	{ "DECOMPOS",   "DECOMPOSE &[] = " },
	{ "DECOMPOSE",  "DECOMPOSE &[] = " },
	{ "DP",         "DEPART " },
	{ "DEP",        "DEPART " },
	{ "DEPA",       "DEPART " },
	{ "DEPAR",      "DEPART " },
	{ "DEPART",     "DEPART " },
	{ "DPS",        "DEPARTS " },
	{ "DEPS",       "DEPARTS " },
	{ "DEPAS",      "DEPARTS " },
	{ "DEPARS",     "DEPARTS " },
	{ "DEPARTS",    "DEPARTS " },
	{ "DPB",        "DEPART &\nBREAK" },
	{ "DEPB",       "DEPART &\nBREAK" },
	{ "DEPAB",      "DEPART &\nBREAK" },
	{ "DEPARB",     "DEPART &\nBREAK" },
	{ "DEPARTB",    "DEPART &\nBREAK" },
	{ "DPSB",       "DEPARTS &" },
	{ "DEPSB",      "DEPARTS &\nBREAK" },
	{ "DEPASB",     "DEPARTS &\nBREAK" },
	{ "DEPARSB",    "DEPARTS &\nBREAK" },
	{ "DEPARTSB",   "DEPARTS &\nBREAK" },
	{ "DET",        "DETACH (&)" },
	{ "DETA",       "DETACH (&)" },
	{ "DETAC",      "DETACH (&)" },
	{ "DETACH",     "DETACH (&)" },
	{ "DIS",        "DISABLE " },
	{ "DISA",       "DISABLE " },
	{ "DISAB",      "DISABLE " },
	{ "DISABL",     "DISABLE " },
	{ "DISABLE",    "DISABLE " },
	{ "DISP",       "DISABLE POWER" },
	{ "DO",         "DO\nUNTIL &" },
	{ "DOS",        "DOS \"&\"" },
	{ "EN",         "ENABLE " },
	{ "ENA",        "ENABLE " },
	{ "ENAB",       "ENABLE " },
	{ "ENABL",      "ENABLE " },
	{ "ENABLE",     "ENABLE " },
	{ "ENP",        "ENABLE POWER" },
	{ "ENV",        "ENABLE VISION" },
	{ "EX",         "EXECUTE &()" },
	{ "EXE",        "EXECUTE &()" },
	{ "EXEC",       "EXECUTE &()" },
	{ "EXECU",      "EXECUTE &()" },
	{ "EXECUT",     "EXECUTE &()" },
	{ "EXECUTE",    "EXECUTE &()" },
	{ "FOR",        "FOR & TO \nEND" },
	{ "FOR;",       "FOR & TO \nEND ; for" },
	{ "GO",         "GOTO " },
	{ "GOT",        "GOTO " },
	{ "GOTO",       "GOTO " },
	{ "IF",         "IF & THEN\nEND" },
	{ "IF;",        "IF & THEN\nEND ; if" },
	{ "IFE",        "IF & THEN\nELSE\nEND" },
	{ "IFEL",       "IF & THEN\nELSE\nEND" },
	{ "IFELS",      "IF & THEN\nELSE\nEND" },
	{ "IFELSE",     "IF & THEN\nELSE\nEND" },
	{ "IFE;",       "IF & THEN\nELSE\nEND ; if/else" },
	{ "IFEL;",      "IF & THEN\nELSE\nEND ; if/else" },
	{ "IFELS;",     "IF & THEN\nELSE\nEND ; if/else" },
	{ "IFELSE;",    "IF & THEN\nELSE\nEND ; if/else" },
	{ "IFG",        "IF & GOTO " },
	{ "L",          "LOCAL " },
	{ "LO",         "LOCAL " },
	{ "LOC",        "LOCAL " },
	{ "LOCAL",      "LOCAL " },
	{ "MC",         "MC " },
	{ "MCS",        "MCS \"&\"" },
	{ "MV",         "MOVE " },
	{ "MOV",        "MOVE " },
	{ "MOVE",       "MOVE " },
	{ "MVB",        "MOVE &\nBREAK" },
	{ "MOVB",       "MOVE &\nBREAK" },
	{ "MOVEB",      "MOVE &\nBREAK" },
	{ "MVS",        "MOVES " },
	{ "MOVS",       "MOVES " },
	{ "MOVES",      "MOVES " },
	{ "MVSB",       "MOVES &\nBREAK" },
	{ "MOVSB",      "MOVES &\nBREAK" },
	{ "MOVESB",     "MOVES &\nBREAK" },
	{ "PAR",        "PARAMETER & = " },
	{ "PARA",       "PARAMETER & = " },
	{ "PARAM",      "PARAMETER & = " },
	{ "PARAME",     "PARAMETER & = " },
	{ "PARAMET",    "PARAMETER & = " },
	{ "PARAMETE",   "PARAMETER & = " },
	{ "PARAMETER",  "PARAMETER & = " },
	{ "PR",         "PROMPT \"&\", " },
	{ "PRO",        "PROMPT \"&\", " },
	{ "PROM",       "PROMPT \"&\", " },
	{ "PROMP",      "PROMPT \"&\", " },
	{ "PROMPT",     "PROMPT \"&\", " },
	{ "REL",        "RELEASE" },
	{ "RELE",       "RELEASE" },
	{ "RELEA",      "RELEASE" },
	{ "RELEAS",     "RELEASE" },
	{ "RELEASE",    "RELEASE" },
	{ "RET",        "RETURN" },
	{ "RETU",       "RETURN" },
	{ "RETUR",      "RETURN" },
	{ "RETURN",     "RETURN" },
	{ "SET",        "SET & = " },
	{ "SI",         "SIGNAL " },
	{ "SIG",        "SIGNAL " },
	{ "SIGN",       "SIGNAL " },
	{ "SIGNA",      "SIGNAL " },
	{ "SIGNAL",     "SIGNAL " },
	{ "SP",         "SPEED " },
	{ "SPE",        "SPEED " },
	{ "SPEE",       "SPEED " },
	{ "SPEED",      "SPEED " },
	{ "SPA",        "SPEED & ALWAYS" },
	{ "TY",         "TYPE " },
	{ "TYP",        "TYPE " },
	{ "TYPE",       "TYPE " },
	{ "VAL",        "VALUE &:" },
	{ "VALU",       "VALUE &:" },
	{ "VALUE",      "VALUE &:" },
	{ "VFA",        "VFIND.ARC(&)" },
	{ "VFIND.A",    "VFIND.ARC(&)" },
	{ "VFIND.AR",   "VFIND.ARC(&)" },
	{ "VFIND.ARC",  "VFIND.ARC(&)" },
	{ "VFL",        "VFIND.LINE(&)" },
	{ "VFIND.L",    "VFIND.LINE(&)" },
	{ "VFIND.LI",   "VFIND.LINE(&)" },
	{ "VFIND.LIN",  "VFIND.LINE(&)" },
	{ "VFIND.LINE", "VFIND.LINE(&)" },
	{ "VFP",        "VFIND.POINT(&)" },
	{ "VFIND.P",    "VFIND.POINT(&)" },
	{ "VFIND.PO",   "VFIND.POINT(&)" },
	{ "VFIND.POI",  "VFIND.POINT(&)" },
	{ "VFIND.POIN", "VFIND.POINT(&)" },
	{ "VFIND.POINT","VFIND.POINT(&)" },
	{ "VLOC",       "VLOCATE(&) " },
	{ "VLOCA",      "VLOCATE(&) " },
	{ "VLOCAT",     "VLOCATE(&) " },
	{ "VLOCATE",    "VLOCATE(&) " },
	{ "VPIC",       "VPICTURE(&)" },
	{ "VPICT",      "VPICTURE(&)" },
	{ "VPICTU",     "VPICTURE(&)" },
	{ "VPICTUR",    "VPICTURE(&)" },
	{ "VPICTURE",   "VPICTURE(&)" },
	{ "VRUL",       "VRULER(&)" },
	{ "VRULE",      "VRULER(&)" },
	{ "VRULER",     "VRULER(&)" },
	{ "W",          "WAIT" },
	{ "WA",         "WAIT" },
	{ "WAI",        "WAIT" },
	{ "WAIT",       "WAIT" },
	{ "W.",         "WAIT.EVENT , " },
	{ "WAIT.",      "WAIT.EVENT , " },
	{ "WH",         "WHILE & DO\nEND" },
	{ "WHI",        "WHILE & DO\nEND" },
	{ "WHIL",       "WHILE & DO\nEND" },
	{ "WHILE",      "WHILE & DO\nEND" },
	{ "WH;",        "WHILE & DO\nEND ; while" },
	{ "WHI;",       "WHILE & DO\nEND ; while" },
	{ "WHIL;",      "WHILE & DO\nEND ; while" },
	{ "WHILE;",     "WHILE & DO\nEND ; while" },
	{ "WR",         "WRITE (&) " },
	{ "WRI",        "WRITE (&) " },
	{ "WRIT",       "WRITE (&) " },
	{ "WRITE",      "WRITE (&) " },
	{ NIL,           NIL }
};

/* Define the list of 'Adept V+' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _pg_names[] = {

"#PDEST",
"#PLATCH",
"#PPOINT",
"$CAL.DATA",
"$CHR",
"$DECODE",
"$DEFAULT",
"$ENCODE",
"$ERROR",
"$FLTB",
"$INTB",
"$LNGB",
"$MID",
"$TIME",
"$TRANSB",
"$TRUNCATE",
"$UNPACK",
".END",
".LOCATIONS",
".PROGRAM",
".REALS",
".STRINGS",
"ABORT",
"ABOVE",
"ABS",
"ACCEL",
"ALIGN",
"ALTER",
"ALTON",
"ALTOFF",
"ALWAYS",
"AMOVE",
"AND",
"ANY",
"APPRO",
"APPROS",
"ASC",
"ATAN2",
"ATTACH",
"AUTO",
"BAND",
"BASE",
"BCD",
"BELOW",
"BELT",
"BELT.MODE",
"BELT.ZERO.COUNT",
"BITS",
"BOR",
"BRAKE",
"BREAK",
"BSTATUS",
"BXOR",
"BY",
"CALIBRATE",
"CALL",
"CALLS",
"CASE",
"CLEAR.EVENT",
"CLOSE",
"CLOSEI",
"COARSE",
"COM",
"CONFIG",
"COS",
"CP",
"CPOFF",
"CPON",
"CYCLE.END",
"D.SCALE.MODE",
"DARC",
"DCB",
"DCLEAR",
"DCURSOR",
"DECOMPOSE",
"DEFBELT",
"DEFINED",
"DELAY",
"DEPART",
"DEPARTS",
"DEST",
"DETACH",
"DEVICE",
"DEVICES",
"DFILL",
"DFLOOD",
"DISABLE",
"DISPLAY.CAMERA",
"DISTANCE",
"DLINE",
"DLINES",
"DO",
"DOS",
"DRIVE",
"DRY.RUN",
"DTYPE",
"DURATION",
"DX",
"DY",
"DZ",
"ELSE",
"ENABLE",
"END",
"ERROR",
"EXECUTE",
"FALSE",
"FCLOSE",
"FCMND",
"FDELETE",
"FEMPTY",
"FINE",
"FLIP",
"FLTB",
"FOPEN",
"FOPENA",
"FOPEND",
"FOPENR",
"FOPENW",
"FOR",
"FORCE",
"FRACT",
"FRAME",
"FREE",
"FSEEK",
"FSET",
"GARC",
"GCHAIN",
"GCLEAR",
"GCLIP",
"GCOLOR",
"GCOPY",
"GET.EVENT",
"GETC",
"GETEVENT",
"GFLOOD",
"GGET.LINE",
"GICON",
"GLINE",
"GLINES",
"GLOGICAL",
"GOTO",
"GPANEL",
"GPOINT",
"GRECTANGLE",
"GSCAN",
"GSLIDE",
"GTEXTURE",
"GTYPE",
"HALT",
"HAND",
"HAND.TIME",
"HERE",
"HOUR.METER",
"ID",
"IDENTICAL",
"IF",
"IGNORE",
"INRANGE",
"INT",
"INTB",
"INTERACTIVE",
"INVERSE",
"IOSTAT",
"IPS","KERMIT.RETRY",
"KERMIT.TIMEOUT",
"KEYMODE",
"KILL",
"LAST",
"LATCH",
"LATCHED",
"LEFTY",
"LEN",
"LNGB",
"LOC",
"LOCAL",
"LOCK",
"MAX",
"MC",
"MCP.MESSAGES",
"MCS",
"MCS.MESSAGES",
"MESSAGES",
"MIN",
"MMPS",
"MOD",
"MOVE",
"MOVEF",
"MOVES",
"MOVESF",
"MOVEST",
"MOVET",
"MULTIPLE",
"NOFILP",
"NONULL",
"NORMAL",
"NOT",
"NOT.CALIBRATED",
"NULL",
"OF",
"OFF",
"ON",
"OPEN",
"OPENI",
"OR",
"PACK",
"PARAMETER",
"PAUSE",
"PAYLOAD",
"PCABORT",
"PCEND",
"PCEXECUTE",
"PCKILL",
"PENDANT",
"PI",
"POS",
"POWER",
"PRIORITY",
"PROMPT",
"RANDOM",
"REACT",
"REACTE",
"REACTI",
"READ",
"READY",
"REAL",
"RELAX",
"RELAXI",
"RELEASE",
"RESET",
"RETRY",
"RETURN",
"RETURNE",
"RIGHTY",
"ROBOT",
"RUNSIG",
"RX",
"RY",
"RZ",
"SCALE",
"SCREEN.TIMEOUT",
"SEE",
"SELECT",
"SET",
"SET.EVENT",
"SET.SPEED",
"SETBELT",
"SETDEVICE",
"SHIFT",
"SIG",
"SIG.INS",
"SIGN",
"SIGNAL",
"SIN",
"SINGLE",
"SOLVE.ANGLES",
"SOLVE.FLAGS",
"SOLVE.TRANS",
"SPEED",
"SQR",
"SQRT",
"STATE",
"STATUS",
"STEP",
"STOP",
"SWITCH",
"TAS",
"TASK",
"TERMINAL",
"THEN",
"TIME",
"TIMER",
"TO",
"TOOL",
"TPS",
"TRACE",
"TRANS",
"TRANSB",
"TRUE",
"TYPE",
"UNTIL",
"UPPER",
"V.2ND.MOMENTS",
"V.2ND.THRESH",
"V.BACKLIGHT"
"V.BINARY",
"V.BORDER.DIST",
"V.BOUNDARIES",
"V.CENTROID",
"V.DISJOINT",
"V.DRY.RUN",
"V.EDGE.INFO",
"V.EDGE.STRENGTH",
"V.EDGE.TYPE",
"V.FIRST.COL",
"V.FIRST.LINE",
"V.FIT.ARCS",
"V.GAIN",
"V.HOLES",
"V.IO.WAIT",
"V.LAST.COL",
"V.LAST.LINE",
"V.LAST.VER.DIST",
"V.MAX.AREA",
"V.MAX.PIXEL.VAR",
"V.MAX.SD",
"V.MAX.TIME",
"V.MAX.VER.DIST",
"V.MIN.AREA",
"V.MIN.HOLE.AREA",
"V.MIN.MAX.RADII",
"V.OFFSET",
"V.OVERLAPPING",
"V.PERIMETER",
"V.RECOGNITION",
"V.SHOW.BOUNDS",
"V.SHOW.EDGES",
"V.SHOW.GRIP",
"V.SHOW.RECOG",
"V.SHOW.VERIFY",
"V.STROBE",
"V.SUBTRACT.HOLE",
"V.SYNCH.STROBE",
"V.THRESHOLD",
"V.TOUCHING",
"VADD",
"VAL",
"VALUE",
"VAUTOTHR",
"VCONVOLVE",
"VCOPY",
"VCORRELATE",
"VDEF.CONVOLVE",
"VDEF.FONT",
"VDEF.MORPH",
"VDEF.SUBRPOTO",
"VDEFGRIP",
"VDELETE",
"VDISPLAY",
"VEDGE",
"VEDGE.INFO",
"VFEATURE",
"VFIND.ARC",
"VFIND.LINE",
"VFIND.POINT",
"VGAPS",
"VGETCAL",
"VGETPIC",
"VHISTOGRAM",
"VISION",
"VLOAD",
"VLOCATE",
"VMORPH",
"VOCR",
"VPICTURE",
"VPUTCAL",
"VPUTPIC",
"VQUEUE",
"VRENAME",
"VRULER",
"VRULERI",
"VSELECT",
"VSHOW",
"VSHOW.MODEL",
"VSTATUS",
"VSTORE",
"VSUBPROTO",
"VSUBTRACT",
"VTHRESHOLD",
"VTRAIN",
"VTRAIN.MODEL",
"VWAIT",
"VWINDOW",
"VWINDOWI",
"WAIT",
"WAIT.EVENT",
"WHILE",
"WINDOW",
"WRITE",
"XOR",

NIL
};



static long _pg_in_comment( long offset );
BOOL _pg_is_firstword( BOOL skip_nonwhite );
UINT _pg_quote_match( void );

void DLL _pg_init( void );
BOOL DLL _pg_expand( void );
void DLL _pg_language_colors_update( long firstline, long lastline );
void DLL _pg_language_colors( long numlines );
void DLL _pg_assign_template( LPSTR abbrev, LPSTR template );
void DLL _pg_add_keyword( LPSTR keyword );
BOOL DLL _pg_indent( void );


void DLL 
_init()
{
	_pg_init();
	CompilerNewExt( ".pg" );
}



/*
 ** _pg_init()
 *
 *  Initialize this module by exporting any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_pg_init( void )
{
	static BOOL firstTime = TRUE;
	int    i;

	if (firstTime)
	{
		LibExport( "_pg_init" );
		LibExport( "BOOL _pg_expand" );
		LibExport( "_pg_language_colors_update long long" );
		LibExport( "_pg_language_colors long" );
		LibExport( "_pg_assign_template LPSTR LPSTR" );
		LibExport( "_pg_add_keyword LPSTR" );
		LibExport( "BOOL _pg_indent( void ) " );
		LibExport( "_pg_routines" );

		firstTime = FALSE;

		/* Create a hash table to hold all of the 'Adept V+' keyword names.
		 */
		_pg_hash_names = HashCreateTable( 41 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _pg_names[ i ] != NIL )
		{
			LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _pg_hash_names, 
					_pg_names[i], strlen( _pg_names[i] ) );
			if (lpf)
				*lpf = StrNew( _pg_names[i] );
			i++;

		}

		/* Create a hash table to hold all of the 'Adept V+' template names.
		 */
		_pg_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _pg_templates[ i ].abbrev != NIL )
		{
			_pg_assign_template( _pg_templates[i].abbrev, _pg_templates[i].template );
			i++;
		}
	}
}


/*
 ** _pg_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'Adept V+' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_pg_expand( void )
{
	LPVOID htEntry;
	LPSTR  _pg_string = NULL;
	LPSTR  str;

	MarkSavePos();

	/* Check to see if we are currently positioned inside an 'Adept V+' comment
	 */
	if (_pg_in_comment( -1L ) != -1)
		goto _pg_expand_incomplete;

	str = BufReadStr( 65535 );  /* read entire line */

	_pg_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_pg_string || !*_pg_string)
		goto _pg_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _pg_hash_templates, _pg_string, strlen( _pg_string ) );
	StrFree( _pg_string );

	if (htEntry)
	{
		/* A template exists for the current keyword, delete the keyword
		 * and insert the template.
		 */
		MarkDropPos( 1 );
		str = *(LPSTR*)htEntry;
		BufDelToEOL();

		_ext_expand_template( str );

	} 
	else
		goto _pg_expand_incomplete;

	return TRUE;

_pg_expand_incomplete:
	MarkRestorePos();

	return FALSE;
}


/*
 ** _pg_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
void DLL
_pg_language_colors_update( long firstline, long lastline )
{
	long startoff;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);
		startoff = _PosQOffset();

		/* Clear all attributes for the line range and
		 * the reset them within _pg_language_colors.
		 */
		AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		_PosInit( startoff );
		_pg_language_colors( lastline - firstline + 1 );
	}
}

/*
 ** _pg_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 *
 * Must include characters that are in variable names but aren't in
 * keywords so that characters that are a part of a variable name but are
 * also a keyword are not colorized.
 * Keywords contain [A-Z].
 * Variable names contain [a-z0-9._].
 * String functions contain [$].
 * Precision points contain [#].
 * Format specifiers for TYPE, WRITE, and $ENCODE contain [/]
 *  "/B","/C","/D","/E","/F","/G","/H","/I","/N","/O","/S","/U","/X"
 *    this would go near the end of the "else if's" below
 *
 *
 */
void DLL
_pg_language_colors( long numlines )
{
	long  startoff;
	long  startline;
	long  endoff;
	long  inc;
	UINT  ch;
	LPSTR buf;
	int   i = 0;
	int   comment_color  = ColorComments( -1 );
	int   keyword_color  = ColorKeywords( -1 );
	int   quote_color    = ColorAlternate1( -1 );
	int   number_color   = ColorAlternate2( -1 );
	int   monitor_color  = ColorAlternate4( -1 );
	int   operator_color = ColorAlternate5( -1 );
	int   linenum_color  = ColorAlternate6( -1 );

	buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 );  
		numlines = 0x7FFFFFFF;
	}
	else
	{
		startoff  = _PosQOffset();
		startline = _PosQLine();
		endoff    = _pg_in_comment( startoff );
		if (endoff == -1)
			_PosInit( startoff );
		else
		{
			_PosInit( endoff );
			numlines += (startline - _PosQLine());
		}
	}
	ch = _PosCurrentChar();
	startoff = _PosQOffset();

	while ((ch != EOF_CHAR) && (numlines > 0))
	{
		if (ch == '\n')  /* if new line character.. */
			numlines--;  /* ..move on to next line */

		/* STRINGS:  look for any characters that are inside quotes denoting a string */
		if (ch == '"')  /* if string open... */
		{
			startoff = _PosQOffset();
			ch = _pg_quote_match();  /* ...close the string */

			/* color the string */
			if (quote_color)
			{
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( quote_color, endoff - startoff );
				_PosInit( endoff );
			}
			continue;
		}

		/* NUMBERS: look for any characters that are part of a number... */
		else if ( (isdigit(ch)) || (ch == '^') || (ch == '\'') )
		{
			UINT lastch;
			UINT whitech;
			BOOL nocolor = FALSE;  /* flag to prevent colorization */
			BOOL exponent;         /* flags more than one 'E' in a float number */
			BOOL radix;
			int numdigits = 0;   /* max # of digits in a number */
			BOOL linenum = FALSE;
			int j;
			char goto_str[] = "GOTO";

			/* look backwards from the current position to find the first
			 * non-space or non-tab character, this character will be used
			 * after the switch statement below to determine whether or
			 * not the number should be colored...
			 *    this is done primarily to prevent line-numbers from
			 *    getting colored, but also to prevent digits that are
			 *    part of an identifier from getting colored */
			startoff = _PosQOffset();  /* save current cursor position */
			lastch = _PosPrevChar();

			if ( isdigit(ch) )
				if ( _pg_is_firstword(FALSE) )
					/* prevent line-numbers (integers preceded only by whitespace on the current line) from getting colored */
					linenum = TRUE;
				else
				{
					/* prevent references to line-numbers (integers preceded by GOTO on the current line) from getting colored */
					do  /* look backwards for non-space or non-tab character */
					{
						whitech = _PosPrevChar();
					} while ( (whitech == ' ') || (whitech == '\t') );  /* while character is whitespace */
					j = 4;
					while ( j > 0 )
					{
						if ( goto_str[--j] != (char)toupper(whitech) )
							j = -1;  /* stop */
						whitech = _PosPrevChar();
					}; /* exits with j==0 if GOTO, otherwise exits with j==-1 */
					linenum = (j == 0);  /* if count reached zero then number is a reference to a line number */
				};

			_PosInit( startoff );  /* restore cursor position */

			switch ( ch )
			{
				/* binary or hexadecimal numbers... */
				case '^':     /* ...as in  ^B0010 (binary)  ^H2FE (hex)  ^73013 (octal)  */
					ch = _PosNextChar();
					numdigits = MAXDIGIT1;  /* initialize digit counter, max # digits after a "^" */
					switch ( toupper(ch) )
					{
						case 'H':   /* hexadecimal numbers */
							do
							{
								ch = _PosNextChar();
							} while ( (isxdigit(ch) || (((ch == '+') || (ch == '-')) && (numdigits == MAXDIGIT1)))
							       && (--numdigits > 0) );
							break;
							
						case 'B':   /* binary numbers */
							do
							{
								ch = _PosNextChar();
							} while ( (((ch == '0') || (ch == '1')) || (((ch == '+') || (ch == '-')) && (numdigits == MAXDIGIT1)))
							       && (--numdigits > 0) );
							break;

						case '0':   /* octal numbers */
						case '1':
						case '2':
						case '3':
						case '4':
						case '5':
						case '6':
						case '7':
						case '+':
						case '-':
							do
							{
								ch = _PosNextChar();
							} while ( (((ch >= '0') && (ch <= '7')) || (((ch == '+') || (ch == '-')) && (numdigits == MAXDIGIT1)))
							       && (--numdigits > 0) );
							break;
						
						default:
							/* don't color anything if [HB+-0-7] don't follow "^" */
							nocolor = TRUE;
							break;
					};
					break;

				/* character constants... */
				case '\'':     /* ...as in  'a  'Z  or  '3 */
					ch = _PosNextChar();
					ch = _PosNextChar();
					numdigits = 1;
					break;
				
				/* integers or floats... */
				default:       /* ...as in 25, -42.0, or 5.39E+27 */
					numdigits = MAXDIGIT2;  /* initialize digit counter, max # digits before the exponent */
					exponent = FALSE;
					radix = FALSE;
					do
					{
						ch = _PosNextChar();

						/* check that there is only one radix... */
						if ( ch == '.' )
						{
							if ( !radix )
							{
								radix = TRUE;
								ch = _PosNextChar();
							}
							else
							{
								nocolor = TRUE;
							};
						};
						
						/* check that there is only one exponent... */
						if ( (ch == 'E') || (ch == 'e') )
						{
							if ( !exponent )
							{
								exponent = TRUE;
								ch = _PosNextChar();
								/* if the character immediatly following the exponent character
								 * is not either a digit or a plus or minus sign, then
								 * do not color the number */
								if ( isdigit(ch) || (ch == '+') || (ch == '-') )
								{
									ch = _PosNextChar();
								}
								else
								{
									nocolor = TRUE;
								};
								numdigits = MAXDIGIT3;  /* re-initialize digit counter, max # digits after an exponent */
							}
							else
							{
								/* don't color anything if more than one 'E' in the number */
								nocolor = TRUE;
							};
						};
					} while ( isdigit(ch) && (--numdigits > 0) );
					break;
			};

			/* don't color number if...   */
			if (( lastch != ' '
			    && lastch != '\t'
			    && lastch != '\r'
			    && lastch != '\n'
			    && lastch != EOF_CHAR    /* ...previous character is not end-of-file   */
				&& !ispunct( lastch ))   /* ...previous character is not punctuation   */
				|| lastch == '_'         /* ...previous character is an under-score    */
				|| nocolor               /* ...flag set to prevent colorization        */
				|| (numdigits <= 0)      /* ...too many digits in the number           */
				|| (ch == '_'))          /* ...next character is an under-score        */
				continue;  /* don't color the number loop if the if-statement evaluates TRUE */

			/* color the number */
			endoff = _PosQOffset();
			_PosInit( startoff );
			if ( linenum )
				_PosSetColor( linenum_color, endoff - startoff );
			else
				_PosSetColor( number_color, endoff - startoff );
			_PosInit( endoff );
		}

		/* NON-KEYWORDS: look for any characters that could be part of a
		 *               keyword or identifier...
		 * 
		 *        This is the set of conditions that will prevent a word from
		 *        falling into a keyword lookup in the last 'else if' statement
		 *        below...
		 *         (i.e. we would color the word "MOVE" if it is surrounded by
		 *               whitespace, but not "$MOVE", "#MOVE",
		 *               or the "MOVE" in "MOVE.FORWARD", "MOVE_FORWARD" */
		else if ( isalnum(ch)
		       || (ch == '.')
		       || (ch == '$')
		       || (ch == '#')
		       || (ch == '_') )
			/* if character is one of [a-zA-Z0-9.$#_] then keep looking... */
		{
			WORD lastch = _PosPrevChar();
	
			if (lastch != EOF_CHAR)
				_PosNextChar();
	
			i = 0;
			startoff = _PosQOffset();

			/* put characters into buffer until character is not one of [a-zA-Z0-9.$#_] */
			do
			{
				if (i < MAX_BUF_LEN)
					buf[i++] = (char) toupper(ch);  /* convert to uppercase for string comparison of "MC" keyword below */
				ch = _PosNextChar();
			} while (isalnum( ch ) || (ch == '.') || (ch == '$') || (ch == '#') || (ch == '_'));
			buf[i] = '\0';  /* null terminate the string */

			/* if leading character of word is not whitespace, punctuation, or
			 * underscore then reset buffer index;
			 * this causes the word to NOT be colorized because the next else
			 * statement will not attempt to look up the word in the keyword
			 * list
			 */
			if ((lastch != ' ' 
			  && lastch != '\t' 
			  && lastch != '\r' 
			  && lastch != '\n' 
			  && lastch != EOF_CHAR
			  && !ispunct( lastch ))
			  || (ch == '_'))
				i = 0;  /* reset buffer index, don't colorize word */

		/* MONITOR COMMAND LINES: look for any characters that are part of an "MC" statement... */
			/* if the first word on the current line is "MC" and the next
			 * non-white space character is not '=' (assignment operator) then
			 * colorize the line as a monitor command line
			 */
			else if ( (strcmp(buf, "MC") == 0) && _pg_is_firstword(TRUE) )
			{
				endoff = _PosQOffset();  /* save current cursor position */
				/* skip over all white space while looking for '=' character */
				while ( ((ch == ' ') || (ch == '\t')) && (ch != EOF_CHAR) )
					ch = _PosNextChar();
				if ( ch != '=' )  /* if next character is not the assignment operator */
				{
					i = 0;  /* reset buffer index, don't colorize word */
					/* once an "MC" keyword is encountered, looks for comment.. */
					do
					{
						ch = _PosNextChar();
						if ( ch == '"' )  /* if string open... */
							ch = _pg_quote_match();  /* ...close the string */
					} while ( (ch != ';') && (ch != '\n') && (ch != EOF_CHAR) );
					endoff = _PosQOffset();
					_PosInit( startoff );
					_PosSetColor( monitor_color, endoff - startoff );
					_PosInit( endoff );
					ch = _PosCurrentChar();
				}
				else
					_PosInit( endoff );  /* restore cursor position */
			};

			continue;
		}

		/* KEYWORDS:  look up the word contained in the buffer (obtained from
		 * previous else statement) in keyword list, colorize if match found
		 */
		else if (i)
		{
			if (HashFindEntry( _pg_hash_names, buf, i ))
			{
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( keyword_color, endoff - startoff );
				_PosInit( endoff );
			}
			i = 0;
		}

		/* COMMENTS: look for any characters that are part of a comment... */
		if (ch == ';')  /* if start of comment... */
		{
			startoff = _PosQOffset();

			/* once a ";" is encountered, skips checking rest of line.. */
			inc = _PosNextLine( 1 ); /* ..move on to next line */
			numlines--;
			endoff = _PosQOffset();
			_PosInit( startoff );
			_PosSetColor( comment_color, endoff - startoff - inc );
			_PosInit( endoff );
			ch = _PosCurrentChar();
			continue;
		}
		
		/* OPERATOR: if character is one of [()[]/+-*<>=,:] */
		if ( (ch == '=')
		       || (ch == '(')
		       || (ch == ')')
		       || (ch == ',')
		       || (ch == '+')
		       || (ch == '-')
		       || (ch == '*')
		       || (ch == '/')
		       || (ch == '<')
		       || (ch == '>')
		       || (ch == '[')
		       || (ch == ']')
		       || (ch == ':') )
		{
			_PosSetColor( operator_color, 1 );
		}

		/* fetch the next character */
		ch = _PosNextChar();
	}
	MemFree( buf );
}



void DLL
_pg_add_keyword( LPSTR keyword )
{
	if (keyword && *keyword)
	{
		LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _pg_hash_names, 
					keyword, strlen( keyword ) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}


/*
 ** _pg_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'Adept V+' template strings
 *    without changing the code.
 */
void DLL
_pg_assign_template( LPSTR abbrev, LPSTR template )
{
	_ext_assign_template( _pg_hash_templates, abbrev, template );
}


/*
 ** _pg_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'Adept V+' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in PG_INDENT_STR - perform smart
 *       indentation.
 */
BOOL DLL
_pg_indent( void )
{
	long   curline;
	long   col;
	LPSTR  _pg_string;
	LPSTR  str;
	long   indent_col;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside an 'Adept V+' comment
	 */
	if (_pg_in_comment( -1L ) != -1)
		goto _pg_indent_incomplete;

	/* Not in a comment so indent according to the current 'Adept V+' construct.
	 */

	indent_col = 0;  /* default value */
	if ( SrchFind( "^\\.PROGRAM", SEARCH_REGEX, NIL )
	  && (BufQCurrentLine() == curline) )
		indent_col = 8;
	else
	{
		MarkRestorePos();
		MarkSavePos();
		if ( SrchFind( "^[ \t]*IF .*GOTO", SEARCH_REGEX, NIL )
		  && (BufQCurrentLine() == curline) )
			goto _pg_indent_incomplete;
		else
		{
			MarkRestorePos();
			MarkSavePos();
			/* If a keyword is not found on the current line then don't indent.
			 */
			/* Place cursor on last character in keyword at beginning of current line. */
			if ( SrchFind( "^[ \t]*[a-zA-Z]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
			  && (BufQCurrentLine() == curline) )
			{
				/* Read the keyword at the beginning of the current line.
				 */
				col = BufQCurrentCol();  /* save cursor position which is last character in keyword as set by call to SrchFind above*/
				/* Place cursor on first character in keyword at beginning of current line. */
				SrchFind( "(^|[^a-zA-Z])\\c[a-zA-Z]", SEARCH_REGEX , NIL );
				str = BufReadStr( col - BufQCurrentCol() );  /* read keyword string from current cursor position to computed length of keyword */
				_pg_string = StrTrim( str, " \t" );
				StrFree( str );
		
				/* Set prefix and suffix to \377 to perform a quick lookup of the name
				 */
				str = StrNew( "\377" );
				_pg_string = StrAppend( _pg_string, str );
				str = StrAppend( str, _pg_string );
				StrFree( _pg_string );
		
				if ( StrMatch( str, PG_INDENT_STR1, SEARCH_IGCASE | SEARCH_FORWARD, NIL ) )
					indent_col = PG_INDENT1;
				else if ( StrMatch( str, PG_INDENT_STR2, SEARCH_IGCASE | SEARCH_FORWARD, NIL ) )
					indent_col = PG_INDENT2;
				StrFree( str );
			};
		};
	};

	if ( indent_col == 0 )
		goto _pg_indent_incomplete;  /* Not an 'Adept V+' keyword, do normal indentation. */

	MarkRestorePos();
	_ext_insert_indented_EOL();  /* indent based on indentation of previous line */
	MovAbs(0, BufQCurrentCol() + indent_col);
	return TRUE;

	_pg_indent_incomplete:
	MarkRestorePos();
	return FALSE;
}



/*--  Private functions --*/


/*
 ** _pg_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'Adept V+' comment by searching backward for a ";"
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_pg_in_comment( long offset )
{
	WORD  ch;

	/* Check to see if we are currently positioned inside an 'Adept V+' comment.
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	do
	{
		if (ch == '\n')
			return -1;
		else if (ch == ';')
			return _PosQOffset();  /* currrent line is in a comment */
		ch = _PosPrevChar();
	} while (ch != EOF_CHAR);

	return -1L;
}


/* _pg_routines()
 * Finds all 'Adept V+' subroutines.
 */

void DLL
_pg_routines( void )
{
	if (Grep("^\\.PROGRAM*"))
		MsgMessage("");
	else
		MsgMessage("No routines found.");
}

/*
 ** _pg_is_firstword()
 *
 *  Determines whether or not the word at the current position is
 *  the first word on the current line by searching backwards for
 *  the new-line character.
 *
 *  Returns -1 if current word is the first word on the current line
 */
BOOL
_pg_is_firstword( BOOL skip_nonwhite )
{
	long startoff;
	UINT ch;

	/* look backwards from the current position to find the first
	 * non-space or non-tab character
	 */
	startoff = _PosQOffset();  /* save current cursor position */

	ch = _PosPrevChar();

	if ( skip_nonwhite )
		while ( isalnum(ch)
		       || (ch == '.')
		       || (ch == '$')
		       || (ch == '#')
		       || (ch == '_') )
			ch = _PosPrevChar();

	while ( (ch == ' ') || (ch == '\t') )  /* while character is whitespace */
		ch = _PosPrevChar();

	_PosInit( startoff );  /* restore cursor position */

	if ( (ch == '\r') || (ch == '\n') || (ch == EOF_CHAR) )
		return TRUE;
	else
		return FALSE;
};

/*
 ** _pg_quote_match()
 *
 *  Steps over characters in buffer until matching quote is found
 *  or until end of line is encountered.  Assumes that leading quote
 *  has already been read into the buffer.
 *
 *  Returns with the buffer position on the next immediate character
 *  from the matching quote and returns that character.
 */
UINT
_pg_quote_match( void )
{
	UINT ch;
	do  /* step over all characters inside the string quotes... */
	{
		ch = _PosNextChar();
	} while ((ch != EOF_CHAR) && (ch != '\n') && (ch != '"'));

	if (ch == '"')
		ch = _PosNextChar();  /* move past closing quote */

	return ch;
};

